/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.sys.testsys.testsys;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.testsys.lib.ArgumentParser;
import de.bsvrz.sys.testsys.lib.MyLogger;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageCauser;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

/**
 * Verwaltet die Überprüfer-Plugins.
 * Der Verwalter startet und verwaltet die Plugins. Er schickt deren Betriebsmeldungen
 * an den DaV.
 *
 * @author Chriesten
 *
 */
public class PluginVerwalter implements IMeldungFehler
{
  /**
   * Ausgabe Logger
   */
  private Debug _debug = Debug.getLogger();


  /**
   * Verbindung zum DaV
   */
  ClientDavInterface m_connection;

  // -------------------------------------------------------------
  // Verwaltung Plugin Instanzen
  // -------------------------------------------------------------
  /**
   * Instanzen aller Plugins
   */
  private List<PluginInst> m_pluginInstList = new ArrayList<PluginInst>();

  // Verwaltung der Plugin - Instanzen, private Klasse.
  private class PluginInst
  {
    PluginInst(Object o, String plugin, String argString)
    {
      m_o = o;
      m_plugin = plugin;
      m_argString = argString;
    }

    // PluginInstanz
    private Object m_o;

    // Plugin Klassennamen
    private String m_plugin;

    // Plugin Argument String
    private String m_argString;

    private void aus()
    {
      MyLogger.logln( "TTT", "plugin="+m_plugin +" ,arg="+m_argString);
    }
  }

  /**
   * Gibt die aktuelle Liste der erzeugten Plugins aus
   */
  public void ausPluginInstList()
  {
    for (PluginInst pi : m_pluginInstList)
    {
      pi.aus();
    }
  }
  // -------------------------------------------------------------

  //---------------------------------------------------------------------------
  // Konstruktoren
  //---------------------------------------------------------------------------
	/**
	 *
	 * @param pluginNamen PluginNamen
	 * @param davConnection DaV
	 */
  public PluginVerwalter(String[] pluginNamen, ClientDavInterface davConnection)
  {
    m_connection = davConnection;
    erzeugePlugins(pluginNamen, davConnection);
  }

  public PluginVerwalter()
  {
  }
  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  // Schnitstellenfunktionen
  //---------------------------------------------------------------------------

  /**
   * Schnittstellenfunktion, siehe {@link IMeldungFehler#fehlerGefunden(String, String)}.
   */
  public void fehlerGefunden( String fehlermeldung, String hinweis )
  {
    String plugin = this.getClass().getName();

    if ( fehlermeldung != null )
      sendMessage (fehlermeldung, 3, plugin);

    if ( hinweis != null )
      sendMessage (hinweis, 1, plugin);
  }


  /**
   * Schnittstellenfunktion, siehe {@link IMeldungFehler#fehlerGefunden(String, Integer, String)}..
   */
  public void fehlerGefunden( String meldung, Integer ebene, String plugin )
  {
    int myebene;
    if ( ebene == null )
      myebene = 1;
    else
      myebene = ebene.intValue();

    sendMessage (meldung, myebene, plugin);
  }
  //---------------------------------------------------------------------------

  /**
   * Erzeugt eine Betriebsmeldung und sendet sie an den
   * Datenverteiler. Eingestellt werden können MeldungsTyp, MeldungsTypZusatz
   * und die Meldungsklasse.
   *
   * @param message Nachricht
   * @param grad Fehlergrad: 1:Info, 2:Warnung, 3:Fehler
   * @param veranlasser Plugin-Bezeichnung
   */
  public void sendMessage (String message, int grad, String veranlasser)
  {
    String id;
    if (m_connection == null)
      id="PluginVerwalter"; //     m_connection.getLocalApplicationObject().getId();
    else
      id=((Long)(m_connection.getLocalApplicationObject().getId())).toString();

    MessageType type = MessageType.APPLICATION_DOMAIN;
    String messageTypeAddOn = null;

    MessageGrade grade;
    switch ( grad )
    {
      case 1:
        grade = MessageGrade.INFORMATION;
      break;

      case 2:
        grade = MessageGrade.WARNING;
      break;

      case 3:
        grade = MessageGrade.ERROR;
      break;

      default:
        grade = MessageGrade.INFORMATION;
      break;
    }

    SystemObject referenceObject = null;
    MessageState state = MessageState.NEW_MESSAGE;
    //MessageCauser causer = new MessageCauser(null, "", "");
    MessageCauser causer = new MessageCauser(null, "", veranlasser);

    MessageSender msg = MessageSender.getInstance();
    msg.setApplicationLabel("PluginVerwalter");

    MyLogger.logln( "BM", "("+grad+")" +message + "(von="+veranlasser+")");
    msg.sendMessage(id, type, messageTypeAddOn, grade, referenceObject, state, causer, message);
  }

  /**
   * Erzeugt aus dem uebergebenen Klassennamen ein Objekt.
   * Falls args != null, wird args im Konstruktor uebergeben
   * Falls argstring != null, wird argstring im Konstruktor uebergeben
   *
   * @param Klasse Klassennamen.
   * @param args Argumente des Klassenkonstruktors
   * @param argstring String alternatives Argument des Klassenkonstruktors
   *
   * @return generiertes Objekt, null im Fehlerfall
   */
  public Object genObject(String Klasse, String[] args, String argstring)
  {
    Object c = null;

    // erwartet der Konstruktor Parameter?
    Boolean konArgsPar = false;
    Boolean konArgstring = false;

    if ( args != null )
    {
      konArgsPar = true;
    }

    if ( argstring != null )
    {
      konArgstring = true;
    }

    try
    {
      Class<?> clazz;
      clazz = (Class<?>)Class.forName(Klasse);

      //--------------------------------------------------------------------
      // Konstruktro mit StringArrayParameter
      //--------------------------------------------------------------------
      if ( konArgsPar )
      {
        try
        {
          // Konstruktor der Klasse mit String Parameter bestimmen
          Class<?>[] paramTypes = new Class[] {String[].class};
          Constructor<?> con=clazz.getConstructor(paramTypes);

          // Neue Instanz der Klasse mit diesem Konstruktor bilden
          c = con.newInstance((Object)args);
        }
        catch ( Exception e )
        {
          String errMess = "Fehler beim Instanzieren mit Parametern. ";
          errMess += "\n" + "Klasse: " + Klasse;
          errMess += "\n" + e.getMessage();
          System.err.println(errMess);

          // Versuch mit anderem Konstrukor
          konArgsPar = false;
        }
      }

      //--------------------------------------------------------------------
      // Konstruktro mit Stringparameter
      //--------------------------------------------------------------------
      if ( (!konArgsPar) && konArgstring )
      {
        try
        {
          // Konstruktor der Klasse mit String Parameter bestimmen
          Class<?>[] paramTypes = new Class[] {String.class};
          Constructor<?> con=clazz.getConstructor(paramTypes);

          // Neue Instanz der Klasse mit diesem Konstruktor bilden
          c = con.newInstance((Object)argstring);
        }
        catch ( Exception e )
        {
          System.err.println("Fehler beim Instanzieren der Klasse " + Klasse +
              "\n mit Parametern ");
          System.out.println(e.getMessage());
          e.printStackTrace();

          // Versuch mit parameterlosem Konstrukor
          konArgstring = false;
        }
      }

      //--------------------------------------------------------------------
      // Konstruktor ohne Parameter
      //--------------------------------------------------------------------
      if ( !konArgsPar && !konArgstring )
      {
        // Instanziierung mit parameterlosem Konstrukor
        Constructor<?> con=clazz.getConstructor((Class[])null);
        c = con.newInstance((Object[])null);
      }

    }
    catch ( Exception e )
    {
      System.out.println("Fehler beim Instanzieren der Klasse " + Klasse);
      System.out.println(e.getMessage());
      e.printStackTrace();
    }
    return c;
  }

  /**
   * Erzeugt Plugin Instanzen. Die Inastanznamen werden aus den übergebenen pluginNamen Zeichenketten
   * generiert. Neben dem Plugin Konstruktors wird die Funktion GibZugriff() gerufen vgl.
   * {@link IGibZugriff#steuereUeberwachung(IMeldungFehler, ClientDavInterface)}.
   * Jede Zeile im pluginNamen Feld enthält den Namen eines Plugins, optional gefolgt von
   * einem ";" und dem Plugin Argumentstring. Ein Argumentstring besteht aus Argumenten der Form <br>
   * key=value. Mehrere Argumente werden durch --- getrennt. <br>
   * Bsp.: de.bsvrz.sys.testsys.modul1:key1=value1---key2=vales2---key3=value3
   * @param pluginNamen siehe obige Beschreibung
   * @param davConnection Aktuelle Verbindung zm DaV
   * @return Anzahl der erfolgreich erzeugten Plugins
   */
  public int erzeugePlugins(String[] pluginNamen, ClientDavInterface davConnection)
  {
    Object o;
    IGibZugriff gibZugriff;
    int erzeugt = 0;
    for (String pluginArg : pluginNamen)
    {
      // Prüfe/Splitte Parameter
      String[] pluginPar = ArgumentParser.splitZk (pluginArg, ";", 2);
      String plugin = pluginPar[0];
      String argString = pluginPar.length==1 ? null : pluginPar[1];

      o = genObject(plugin, null, argString);
      if ( o != null)
      {
        if ( !(o instanceof IGibZugriff) )
        { // Dieses Interface muss implementiert sein
          System.err.println("Die Klasse " + plugin + " muss die Schnittstelle GibZugriff implemenieren!");
          o = null;
          continue;
        }

        gibZugriff = (IGibZugriff)o;
        // Plugin speichern
        MyLogger.logln( "VERLAUF", "Plugin erzeugt: "+o.getClass().getName());
        gibZugriff.steuereUeberwachung( (IMeldungFehler)this, davConnection);

        // Plugin merken
        m_pluginInstList.add( new PluginInst(o, plugin, argString) );
        erzeugt++;
      }
    }
    return(erzeugt);
  }

  /**
   * Beendet alle Plugin Instanz mit dem übergebenen Namen, falls das
   * Plugin die Schnittstelle {@link IBeendePlugin} implementiert.
   * @param plugin Plugin
   */
  public void beendePlugin(String plugin)
  {
    // optionalen Argumentstring entfernen
    int pos = plugin.indexOf( ":" );
    if ( pos != -1 )
      plugin = plugin.substring( 0, pos );

    for (PluginInst pi : m_pluginInstList)
    {
      if ( (pi.m_plugin.equalsIgnoreCase( plugin )) )
      {
        Object o = pi.m_o;
        if ( o instanceof IBeendePlugin )
        {
          IBeendePlugin beendePlugin  = (IBeendePlugin)o;
          beendePlugin.beenden();
        };
      }
    }
  }

  /**
   * Beendet alle vom Verwalter gestarteten Plugin, die die 
   * Schnittstelle {@link IBeendePlugin} implementieren.
   */
  public void beendePlugins()
  {
    for (Iterator<PluginInst> it = m_pluginInstList.iterator(); it.hasNext();)
    {
      PluginInst pi = (PluginInst) it.next();
      Object o = pi.m_o;
      if ( o instanceof IBeendePlugin )
      {
        IBeendePlugin beendePlugin  = (IBeendePlugin)o;
        beendePlugin.beenden();
      };
    }
  }
  
  /**
   * Testroutine fuer das lokale Modul. Wird fuer verschiedene Tests im Laufe des Entwicklungs- und Wartungsprozesses verwendet
   * 
   * @param args
   *          Testargumente
   */
  public static void main( String[] args )
  {
    String[] pluginNamen = new String[1];
    pluginNamen[0] = "de.bsvrz.sys.testsys.modulUhrzeit.ModulUhrzeit";
    pluginNamen[0] = "de.bsvrz.sys.testsys.modulStartStoppUeb.ModulStartStoppUeb:rechner=konfig";
    pluginNamen[0] = "de.bsvrz.sys.testsys.modulApplUeb.ModulApplUeb:rechner=lokal";
    PluginVerwalter verw = new PluginVerwalter(pluginNamen, null);
    if ( verw == null ) System.out.println("kann PluginVerwalter() nicht erzeugen");
if (true) return;
    Object o;
    String plugin = "de.bsvrz.sys.testsys.modulUhrzeit.ModulUhrzeit";
    String[] myargs = new String[2];
    myargs[0] = "000";
    myargs[1] = "111";
    PluginVerwalter pv = new PluginVerwalter();
    o = pv.genObject(plugin, null, null);
    pv = null;
    if ( true)
      return;
  }
}
