/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.sys.testsys.testsys.davlib;


import java.net.InetAddress;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObjectType;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.testsys.lib.MyLogger;
import de.bsvrz.sys.testsys.testsys.UeberpruefungSystemApp;

public class DavLib implements ClientReceiverInterface, ClientSenderInterface, IDaVInfoAnbieter
{
  private Debug _debug = Debug.getLogger();

  /**
   * Aktuelle Verbindung zum Datenverteiler.
   */
  private ClientDavInterface m_connection;

  /**
   * Liste aller Objekte, die am update des Aspekts recher.ProzInfo.Zustand interessiert sind
   */
  private HashMap<IDaVInfoKonsument, IDaVInfoKonsument> m_prozInfoKonsumenten = new HashMap<IDaVInfoKonsument, IDaVInfoKonsument>();


  // ---------------------------------------------
  // Konstruktor
  // ---------------------------------------------
  public DavLib( ClientDavInterface connection )
  {
    m_connection = connection;
  }

  // ---------------------------------------------

  // -----------------------------------------------------------------------

  //------------------------------------------------
  // Methoden der Schnittstelle IDaVInfoAnbieter
  //------------------------------------------------
  /**
   * siehe {@link IDaVInfoAnbieter}{@link #abonniereDaVInfo(IDaVInfoKonsument)}
   */
  public boolean abonniereDaVInfo( IDaVInfoKonsument kon )
  {
    if ( m_prozInfoKonsumenten.containsKey( kon ) )
      return (false);

    m_prozInfoKonsumenten.put( kon, kon );

    return (true);
  }

  /**
   * siehe {@link IDaVInfoAnbieter}{@link #kuendigeDaVInfo(IDaVInfoKonsument)}
   */
  public boolean kuendigeDaVInfo( IDaVInfoKonsument kon )
  {
    if ( m_prozInfoKonsumenten.containsKey( kon ) )
    {
      m_prozInfoKonsumenten.remove( kon );
      return (true);
    }
    else
      return (false);
  }

  /**
   * Uebergibt die aktuelle DAV Information an alle interessierten Konsumenten
   * @param dat Ergebnisdatum vom DAV
   */
  private void uebergebeNeueDavInfo( ResultData dat )
  {
    if ( m_prozInfoKonsumenten == null )
      return;
    
    for (Iterator<IDaVInfoKonsument> it = m_prozInfoKonsumenten.values().iterator(); it.hasNext();)
    {
      IDaVInfoKonsument kon = (IDaVInfoKonsument) it.next();
      kon.neueDavInfo( dat);
    }
  }
//------------------------------------------------


  /**
   * Ergänzt übergebene Rechnerliste um die am
   * aktuellen Datenverteiler konfigurierten Rechnerobjekte == Aspekt typ.rechner.rechnerInformation.Eigenschaft
   * @param rdlist Map
   * @param rechner null,lokal-&gt;nur lokalen (aktuellen) Rechner, konfig-&gt;alle am DAV konfigurierten Rechner
   * sonst-&gt;den uebergebenen Rechner
   * @param connection
   *          Verbindung zum DAV
   */
  // -----------------------------------------------------------------------
  public void leseRechnerInformation(HashMap<String, RechnerDaten> rdlist, String rechner, ClientDavInterface connection )
  {
    MyLogger.logln( "VERLAUF", this.getClass().getName() + ".leseRechnerInformation("+rechner+","+connection+") aus DAV" );

    Boolean alleRechner = false;

    SystemObjectType typeSysObj = null;

    // aktuellen Rechner ermittlen
    String rechner_lokal = "";
    String tcpip_lokal = "";

    try
    {
      InetAddress thisInetAddress = InetAddress.getLocalHost();
      rechner_lokal = thisInetAddress.getHostName();
      tcpip_lokal = thisInetAddress.getHostAddress();

      String localHost = InetAddress.getLocalHost().getHostName();
      InetAddress[] ia = InetAddress.getAllByName(localHost);
    }
    catch ( Exception e )
    {
      String errMess = "InetAddress.getLocalHost().getHostName() Fehler" ;
      errMess += "\n" + e.getMessage();
      MyLogger.logln( "ERROR", errMess );
      _debug.error( errMess );
      e.printStackTrace();
    }

    if ( rechner != null )
    {
      tcpip_lokal = rechner;
      rechner_lokal = rechner;
    }

    String mess = "";
    mess += "rechner_lokal="+rechner_lokal;
    mess += " ,tcpip_lokal="+tcpip_lokal;
System.out.println(mess);


    if ( connection == null )
    {
      MyLogger.logln( "ERROR", "keine Verbindung zum DAV" );
      return;
    }

    DataModel dataModel = connection.getDataModel();

    typeSysObj = dataModel.getType( "typ.rechner" );
    if ( typeSysObj == null )
    {
      MyLogger.logln( "ERROR", "typ.rechner unbekannt" );
      System.err.println( "typ.rechner unbekannt" );
      _debug.error( "typ.rechner unbekannt" );
      return;
    }

    if ( typeSysObj.isConfigurating() )
      typeSysObj = (ConfigurationObjectType) typeSysObj;
    else
      typeSysObj = (DynamicObjectType) typeSysObj;

    // Elemente dieses Systemobjekts einlesen
    List<SystemObject> typobjekte = typeSysObj.getElements();

    Iterator<SystemObject> iterator_Typen = typobjekte.iterator();

    while ( iterator_Typen.hasNext() )
    {
      SystemObject sysObj = (SystemObject) iterator_Typen.next();
      if ( sysObj != null )
      {
        Data data = sysObj.getConfigurationData( dataModel.getAttributeGroup( "atg.rechnerInformation" ) );

        if ( data == null )
        {
          MyLogger.logln( "ERROR", "Für Rechner <" + sysObj.getName()
              + "> ist Attributgruppe <atg.rechnerInformation> nicht definiert" );
        }
        else
        {
          RechnerDaten rechnerdaten = new RechnerDaten();

          rechnerdaten.m_rechnerPid = sysObj.getPid();
          rechnerdaten.m_rechnerObjektName = sysObj.getName();
          rechnerdaten.m_name = data.getItem( "Name" ).valueToString();
          rechnerdaten.m_tcpip = data.getItem( "TCPIP" ).valueToString();
          if ( rechnerdaten.m_tcpip.equalsIgnoreCase( tcpip_lokal ) )
            rechnerdaten.m_ist_aktuell = true;
          else if ( rechnerdaten.m_rechnerPid.equalsIgnoreCase( rechner_lokal ) )
            rechnerdaten.m_ist_aktuell = true;
          else if ( rechnerdaten.m_rechnerObjektName.equalsIgnoreCase( rechner_lokal ) )
            rechnerdaten.m_ist_aktuell = true;
          else if ( rechnerdaten.m_name.equalsIgnoreCase( rechner_lokal ) )
            rechnerdaten.m_ist_aktuell = true;

          if ( alleRechner || rechnerdaten.m_ist_aktuell)
            rdlist.put( rechnerdaten.m_rechnerPid, rechnerdaten );
        }
      }
    } // while


    if ( rdlist.size() == 0 )
    {
      MyLogger.logln( "ERROR", "Keine Konfiguration für Rechner "+rechner_lokal+"("+tcpip_lokal+").");
      (new Throwable()).printStackTrace(MyLogger.get_out());
    }
    return;
  }

  // -----------------------------------------------------------------------

  /**
   * Meldet alle Rechnerelemente aus rdl am DaV an.
   * Angemeldete wird jeweils typ.rechner.prozessInfo.zustand mit der ObjPid RechnerDaten.m_rechnerPid.
   * @param rdl siehe obige Beschreibung
   * @param rechner siehe 2. Parameter in {@link DavLib}{@link #leseRechnerInformation(HashMap, String, ClientDavInterface)}
   */
  public void anmeldeLeseProzessInfoList( HashMap<String, RechnerDaten> rdl,String rechner )
  {
    leseRechnerInformation(rdl, rechner, m_connection);
    for (Iterator<RechnerDaten> it = rdl.values().iterator(); it.hasNext();)
    {
      RechnerDaten rd = (RechnerDaten) it.next();
      anmeldeLeseProzessInfo( rd.m_rechnerPid );
    }
    return;
  }

  /**
   * Meldet alle Rechnerobjekte in rdl von Lese Aspekt "typ.rechner.prozessInfo.zustand" ab.
   * @param rdl Liste der auf prozessInfo angemeldeten Rechner
   */
  public void abmeldeLeseProzessInfoAuto( HashMap<String, RechnerDaten> rdl)
  {
    for (Iterator<RechnerDaten> it = rdl.values().iterator(); it.hasNext();)
    {
      RechnerDaten rd = (RechnerDaten) it.next();
      abmeldeLeseProzessInfo( rd.m_rechnerPid );
    }
  }


  /**
   * Methode zur DaV Empfangs Anmeldung für Rechners Objekt, Attributgruppe typ.rechner.prozessInfo.zustand.
   *
   * @param pidRechner
   *          pid des Objekts vom Type typ.rechner
   * @return boolean
   */
  public boolean anmeldeLeseProzessInfo( String pidRechner )
  {
    boolean bret = false;

    String objPid = pidRechner;
    String atgPid = "atg.prozessInfo";
    String aspPid = "asp.zustand";
    ReceiverRole role = ReceiverRole.receiver();

    bret = anmeldeLeseDav( objPid, atgPid, aspPid, role );
    return (bret);
  }

  /**
   * Methode zur DaV Empfang Abmeldung für Rechners Objekt, Attributgruppe typ.rechner.prozessInfo.zustand.
   *
   * @param pidRechner
   *          pid des Objekts vom Type typ.rechner
   */
  public void abmeldeLeseProzessInfo( String pidRechner )
  {
    String objPid = pidRechner;
    String atgPid = "atg.prozessInfo";
    String aspPid = "asp.zustand";

    abmeldeLeseDav( objPid, atgPid, aspPid );
    return;
  }

  /**
   * Methode zur DaV Empfangs - Anmeldung für Objekt vom Type typ.datenverteiler, Attributgruppe
   * angemeldeteApplikationen
   *
   * @param pidDAV
   *          pid des Objekts vom Type typ.datenverteiler
   * @return boolean
   */
  public boolean anmeldeLeseAngemeldeteAppl( String pidDAV )
  {
    boolean bret = false;

    if ( pidDAV == null )
    {
      pidDAV = "dav.nw.leverkusen.nba.koko";
      pidDAV = "LocalDav";
    }

    String objPid = pidDAV;
    String atgPid = "atg.angemeldeteApplikationen";
    String aspPid = "asp.standard";
    ReceiverRole role = ReceiverRole.receiver();

    bret = anmeldeLeseDav( objPid, atgPid, aspPid, role );
    return (bret);
  }

  /**
   * Methode zur DaV Empfangs - Abmeldung für Objekt vom Type typ.datenverteiler, Attributgruppe
   *
   * @param pidDAV
   *          pid des Objekts vom Type typ.datenverteiler
   */
  public void abmeldeLeseAngemeldeteAppl( String pidDAV )
  {
    if ( pidDAV == null )
    {
      pidDAV = "dav.nw.leverkusen.nba.koko";
    }

    String objPid = pidDAV;
    String atgPid = "atg.angemeldeteApplikationen";
    String aspPid = "asp.standard";

    abmeldeLeseDav( objPid, atgPid, aspPid );
    return;
  }

  /**
   * Methode zur DaV Empfangs Anmeldung für Objekte vom Type typ.applikation ; Attributgruppe
   * ApplikationsFertigmedlung.
   *
   * @param applId
   *          Id des Objekts vom Type typ.appliaktion
   * @return boolean
   */
  public boolean anmeldeLeseApplFertig( String applId )
  {
    boolean bret = false;

    String atgPid = "atg.applikationsFertigmeldung";
    String aspPid = "asp.standard";
    ReceiverRole role = ReceiverRole.receiver();

    bret = anmeldeLeseDav( applId, atgPid, aspPid, role );
    return (bret);
  }

  /**
   * Methode zur DaV Empfangs Abmeldung für Objekte vom Type typ.applikation ; Attributgruppe
   * ApplikationsFertigmedlung
   *
   * @param applId
   *          Id des Objekts vom Type typ.appliaktion
   */
  public void abmeldeLeseApplFertig( String applId )
  {
    String objPid = applId;
    String atgPid = "atg.applikationsFertigmeldung";
    String aspPid = "asp.standard";

    abmeldeLeseDav( objPid, atgPid, aspPid );
    return;
  }

  /**
   * Methode die das Objekt mit der PID objPid beim Datenverteiler zum Lesen anmeldet,
   * Attibutgruppe atgPid unter dem Aspekt aspPid.
   *
   * @param objPid
   *          Pid des Objekts
   * @param atgPid
   *          Attributgruppe die angemeldet werden soll
   * @param aspPid
   *          Apekt der angemeldet werden soll
   * @param role
   *          Rolle des Empfängers (siehe stauma.dav.clientside.ReceiveOptions)
   * @return true: ok, false: nicht ok
   */
  private boolean anmeldeLeseDav( String objPid, String atgPid, String aspPid, ReceiverRole role )
  {
    String buffer = "Anmelden am DaV (Lesen): " + objPid + " " + atgPid + "  " + aspPid;
    MyLogger.logln( "VERLAUF", buffer );

    ClientDavInterface verb = m_connection;

    try
    {
      AttributeGroup atg = verb.getDataModel().getAttributeGroup( atgPid );
      Aspect asp = verb.getDataModel().getAspect( aspPid );

      DataDescription dd = new DataDescription( atg, asp );

      SystemObject obj;
      if ( objPid.matches( "[0-9]*" ) )
      { // nur Zahlen == id
          Long objId = new Long(objPid );
          obj = verb.getDataModel().getObject( objId );
      }
      else if ( objPid.equalsIgnoreCase( "LocalDav") )
      {
        obj = m_connection.getLocalDav();
      }
      else // pid
      {
        obj = verb.getDataModel().getObject( objPid );
      }

      verb.subscribeReceiver( this, obj, dd, ReceiveOptions.normal(), role );
    }
    catch ( Exception e1 )
    {
      MyLogger.logln( "ERROR", "subscribeReceiver:" + buffer );
      MyLogger.logln( "ERROR", e1.getMessage() );
      MyLogger.log( "ERROR", "" );
      return (false);
    }
    return (true);
  }

  /**
   * Methode die das Objekt mit der PID objPid beim Datenverteiler vom Lesen abmeldet, 
   * Attibutgruppe atgPid unter dem Aspekt aspPid.
   * 
   * @param objPid
   *          Pid oder Id des Objekts
   * @param atgPid
   *          Attributgruppe die abgemeldet werden soll
   * @param aspPid
   *          Apekt der abgemeldet werden soll
   */
  public void abmeldeLeseDav( String objPid, String atgPid, String aspPid )
  {
    String buffer = "Abmelden vom DaV (Lesen): " + objPid + " " + atgPid + "  " + aspPid;
    MyLogger.logln( "VERLAUF", buffer );

    ClientDavInterface verb = m_connection;

    try
    {
      AttributeGroup atg = verb.getDataModel().getAttributeGroup( atgPid );
      Aspect asp = verb.getDataModel().getAspect( aspPid );

      DataDescription dd = new DataDescription( atg, asp );

      SystemObject obj;
      if ( objPid.matches( "[0-9]*" ) )
      { // nur Zahlen == id
          Long objId = new Long(objPid );
          obj = verb.getDataModel().getObject( objId );
      }
      else // pid
      {
        obj = verb.getDataModel().getObject( objPid );
      }

      verb.unsubscribeReceiver( this, obj, dd );
    }
    catch ( Exception e1 )
    {
      MyLogger.logln( "ERROR", "unsubscribeReceiver:" + buffer );
      MyLogger.logln( "ERROR", e1.getMessage() );
      e1.printStackTrace();
    }
  }

  /**
   * Methode die das Objekt mit der PID objPid beim Datenverteiler zum senden anmeldet, 
   * Attibutgruppe atgPid, Aspekt aspPid.

   * @param objPid
   *          Pid des Objekts
   * @param atgPid
   *          Pid der Attributgruppe
   * @param aspPid
   *          Pid des Aspekts
   * @param role Role
   * @return DataDescription
   */
  public DataDescription anmeldeSendeDav( String objPid, String atgPid, String aspPid, SenderRole role )
  {
    String buffer = "Anmelden am DaV (Senden): " + objPid + " " + atgPid + "  " + aspPid + " " + role.toString();
    MyLogger.logln( "VERLAUF", buffer );

    ClientDavInterface verb = m_connection;

    DataDescription dd = null;
    try
    {
      AttributeGroup atg = verb.getDataModel().getAttributeGroup( atgPid );
      Aspect asp = verb.getDataModel().getAspect( aspPid );

      dd = new DataDescription( atg, asp );

      SystemObject sysObj = verb.getDataModel().getObject( objPid );

      verb.subscribeSender( this, sysObj, dd, SenderRole.source() );
    }
    catch ( Exception e1 )
    {
      MyLogger.logln( "ERROR", "subscribeSender:" + buffer );
      MyLogger.logln( "ERROR", e1.getMessage() );
      MyLogger.log( "ERROR", "" );
      e1.printStackTrace( MyLogger.get_out() );
      e1.printStackTrace();
    }
    return dd;
  }

  /**
   * Methode die das Objekt mit der PID objPid beim Datenverteiler vom senden abmeldet, 
   * Attibutgruppe atgPid, Aspekt aspPid.
   * 
   * @param objPid
   *          Pid des Objekts
   * @param atgPid
   *          Attributgruppe die abgemeldet werden soll
   * @param aspPid
   *          Apekt der abgemeldet werden soll
   */
  public void abmeldeSendeDaV( String objPid, String atgPid, String aspPid )
  {
    String buffer = "Abmelden vom DaV (Senden): " + objPid + " " + atgPid + "  " + aspPid;
    MyLogger.logln( "VERLAUF", buffer );

    ClientDavInterface verb = m_connection;

    try
    {
      AttributeGroup atg = verb.getDataModel().getAttributeGroup( atgPid );
      Aspect asp = verb.getDataModel().getAspect( aspPid );

      DataDescription dd = new DataDescription( atg, asp );

      SystemObject sysObj = verb.getDataModel().getObject( objPid );

      verb.unsubscribeSender( this, sysObj, dd );
    }
    catch ( Exception e1 )
    {
      MyLogger.logln( "ERROR", "unsubscribeSender:" + buffer );
      MyLogger.logln( "ERROR", e1.getMessage() );
    }
  }

  // -------------------------------------------------
  // Schnittstelle ClientSenderInterface
  // -------------------------------------------------
  /**
   * see {@link ClientSenderInterface}{@link #dataRequest(SystemObject, DataDescription, byte)}
   */
  public void dataRequest( SystemObject object, DataDescription dataDescription, byte state )
  {
  }

  /**
   * see {@link ClientSenderInterface}{@link #isRequestSupported(SystemObject, DataDescription)}
   */
  public boolean isRequestSupported( SystemObject object, DataDescription dataDescription )
  {
    return false;
  }

  // -------------------------------------------------

  // -------------------------------------------------
  // Schnittstelle ClientReceiverInterface
  // -------------------------------------------------
  /**
   * see {@link ClientReceiverInterface}{@link #update(ResultData[])}
   */
  public void update( ResultData[] results )
  {
    for (int i = 0; i < results.length; i++)
    {
      ResultData dat = results[i];

      uebergebeNeueDavInfo( dat );
      
      if ( dat.getDataDescription().getAttributeGroup().getPid().equals( "atg.prozessInfo" ) )
      {
        MyLogger.logln( "TTT_UPD", "atg.prozessInfo" );
        // updateProzessInfo( dat );
      }
      else if ( dat.getDataDescription().getAttributeGroup().getPid().equals( "atg.angemeldeteApplikationen" ) )
      {
        // updateAngemeldeteAppl( dat );
        MyLogger.logln( "TTT_UPD", "atg.angemeldeteApplikationen" );
      }
      else if ( dat.getDataDescription().getAttributeGroup().getPid().equals( "atg.applikationsFertigmeldung" ) )
      {
        // updateApplFertig( dat );
        MyLogger.logln( "TTT_UPD", "atg.applikationsFertigmeldung" );
      }
    }
  }
  // -------------------------------------------------

  
}
